@@ -24,6 +24,10 @@
 # sys = nf(freq, zeta, depth, dt=None, method='matched')
 # sys = pf(freq, zeta, k, phi, dt=None, method='matched')
 # sys = pfopt(freq, zeta, kpdb, sysT, dt=None, method='matched')
+# DOBu, DOBy = dob(freq, zeta, M, C, K, dt, nd = 0)
+# Czpetc, Nzpetc = zpetc(Pz, zerothr=0.99)
+# sys = filt(num, den, dt)
+# sys = minreal(sys)
 
 
 import math
@@ -300,6 +304,8 @@ def zpk2ss(zpk, form='reachable'):
 def sys2frd(sys, freq):
     if type(sys) == ZpkModel:
         sys = zpk2tf(sys)
+    if type(freq) == list:
+        freq = np.array(freq)
     mag, phase, omega = matlab.freqresp(sys, freq*2.0*np.pi)
     real = mag*np.cos(phase)
     imag = mag*np.sin(phase)
@@ -513,20 +519,15 @@ def _pfoptparam(freq, zeta, depth, sysT):
     if (len(freq)==len(zeta)==len(depth)) == False:
         print('Error: length of peak filter parameters is different!')
     omega = 2.0*np.pi*np.array(freq)
-    if type(sysT) == FreqResp:
-        re_invT = np.array([])
-        im_invT = np.array([])
-        invsysT = 1.0/sysT
-        for i in range(len(freq)):
-            idx = np.argmin(np.abs(invsysT.freq - freq[i]))
-            re_invT = np.append(re_invT, np.real(invsysT.resp[idx]))
-            im_invT = np.append(im_invT, np.imag(invsysT.resp[idx]))
-    else:
-        if type(sysT) == ZpkModel:
-            sysT = zpk2tf(sysT)
-        mag, phase, tmp = matlab.freqresp(1.0 / sysT, omega)
-        re_invT = mag * np.cos(phase)
-        im_invT = mag * np.sin(phase)
+    invsysT = 1.0 / sysT
+    if type(invsysT) != FreqResp:
+        invsysT = sys2frd(invsysT, np.array(freq))
+    re_invT = np.array([])
+    im_invT = np.array([])
+    for i in range(len(freq)):
+        idx = np.argmin(np.abs(invsysT.freq - freq[i]))
+        re_invT = np.append(re_invT, np.real(invsysT.resp[idx]))
+        im_invT = np.append(im_invT, np.imag(invsysT.resp[idx]))
 
     norm_invT = np.sqrt(re_invT**2+im_invT**2)
     phi = -omega*re_invT/im_invT
@@ -539,3 +540,124 @@ def _pfoptparam(freq, zeta, depth, sysT):
         if phaseT/phaseF < 0:
             k[i] *= -1
     return freq, zeta, k, phi
+
+
+def dob(freq, zeta, M, C, K, dt, nd = 0):
+    # d_hat = -Q[z] * u + Q[z]*P[z]**-1 * y
+    # DOBu = Q[z], DOBy = Q[z]*P[z]
+
+    # continuous-time plant model
+    Acp = np.matrix([[0.0, 1.0], [-K/M, -C/M]])
+    bcp = np.matrix([[0.0], [1/M]])
+    ccp = np.matrix([[1.0, 0.0]])
+    Pc = ss(Acp, bcp, ccp, 0)
+    # discrete-time plant model
+    Ps = c2d(Pc, dt, method='zoh')
+    Asp = Ps.A
+    bsp = Ps.B
+    csp = Ps.C
+
+    # discrete-time disturbance model
+    Asd = np.matrix([[1.0]])
+    csd = np.matrix([[1.0]])
+
+    # augumented-discrete time plant model
+    As = np.append(np.append(Asp, bsp*csd, 1), [np.append(np.zeros([1, 2]), Asd)], 0)
+    bs = np.append(bsp, np.zeros([1, 1]), 0)
+
+    # Min-order observer
+    Aob11 = As[0, 0]
+    Aob12 = As[0,1:3]
+    Aob21 = As[1:3, 0]
+    Aob22 = As[1:3, 1:3]
+    bob1 = bs[0]
+    bob2 = bs[1:3]
+
+    # pole placement
+    if zeta >= 1.0:
+        w1 = -2.0 * np.pi * freq * (zeta + np.sqrt(zeta ** 2 - 1))
+        w2 = -2.0 * np.pi * freq * (zeta - np.sqrt(zeta ** 2 - 1))
+    else:
+        w1 = -2.0 * np.pi * freq * (zeta + np.sqrt(-zeta ** 2 + 1) * 1.j)
+        w2 = -2.0 * np.pi * freq * (zeta - np.sqrt(-zeta ** 2 + 1) * 1.j)
+    pole_obs = np.array([np.exp(w1*dt), np.exp(w2*dt)])
+    K = matlab.acker(Aob22.T, Aob12.T, pole_obs).T
+
+    # observer matrices
+    Ahat = Aob22 - K * Aob12
+    bhat = Aob21 + Aob22 * K - K * Aob11 - K * Aob12 * K
+    Jhat = bob2 - K * bob1
+    Chat = np.append(np.zeros([1, 2]), np.eye(2), 0)
+    dhat = np.append(np.array([[1.0]]), K, 0)
+
+    # DOB
+    Adob = Ahat
+    Bdob = np.append(bhat, Jhat, 1)
+    Cdob = np.append(np.zeros([1, 2]), csd, 1) * Chat
+    Ddob = np.append(np.append(np.zeros([1, 2]), csd, 1) * dhat, np.zeros([1, 1]), 1)
+    # TransferFunction
+    DOBy = ss2tf(ss(Adob, Bdob[:, 0], Cdob, Ddob[:, 0], dt))
+    DOBu = -ss2tf(ss(Adob, Bdob[:, 1], Cdob, Ddob[:, 1], dt))
+
+    if nd > 0:
+        DOBu_num, DOBu_den = matlab.tfdata(DOBu)
+        DOBu_den = np.append(DOBu_den[0], np.zeros([1, int(nd)]), 1)
+        DOBu = tf(DOBu_num[0][0], DOBu_den[0], dt)
+
+    return DOBu, DOBy
+
+
+def zpetc(Pz, zerothr=0.99):
+    denz = Pz.den[0][0]
+    B_zero = tf2zpk(Pz).z
+    A_pole = tf2zpk(Pz).p
+    B_gain = tf2zpk(Pz).k
+
+    mNbr = len(B_zero)          # Number of zeros
+    nNbr = len(A_pole)          # Number of poles
+    dNbr = nNbr - mNbr          # z^(-d) for b0 not to be zero
+
+    B_UnSt_zero = []
+    B_St_zero = []
+    for i in range(mNbr):
+        if abs(B_zero[i]) >= zerothr:
+            B_UnSt_zero.append(B_zero[i])
+        else:
+            B_St_zero.append(B_zero[i])
+    sNbr = len(B_UnSt_zero)     # Number of unstable zeros
+
+    # Polynomial of unstable zeros
+    B_UnSt_z1 = polynomial.polyfromroots(B_UnSt_zero) * B_gain
+    # Polynomial of stable zeros
+    B_St_z1 = polynomial.polyfromroots(B_St_zero)
+    B_UnStAs_z1 = np.flip(B_UnSt_z1)
+    # DC gain of polynomial of unstable zeros
+    B_UnSt_1 = np.sum(B_UnSt_z1)
+    # Number of forward samples
+    Nzpetc = dNbr + sNbr
+
+    # Transfer function of ZPETC
+    nz_z = np.convolve(denz, B_UnStAs_z1)
+    dz_z = np.convolve(B_St_z1, B_UnSt_1**2)
+    Czpetc = filt(nz_z, dz_z, Pz.dt)
+    return Czpetc, Nzpetc
+
+
+def filt(num, den, dt):
+    zinv = tf([1], [1, 0], dt)
+    numpoly = 0
+    for i in range(len(num)):
+        numpoly += num[i] * zinv**i
+    denpoly = 0
+    for i in range(len(den)):
+        denpoly += den[i] * zinv**i
+    return minreal(numpoly/denpoly)
+
+
+def minreal(sys):
+    if type(sys) == matlab.TransferFunction:
+        return zpk2tf(tf2zpk(sys))
+    elif type(sys) == matlab.StateSpace:
+        return zpk2ss(ss2zpk(sys))
+    else:
+        return sys
